from typing import List
from sqlalchemy import select
from sqlalchemy.orm import joinedload

from open_aoi.controllers import Controller
from open_aoi.models import (
    ControlHandlerModel,
    DefectTypeModel,
    ControlTargetModel,
)


class ControlHandlerController(Controller):
    _model = ControlHandlerModel

    def create(
        self, title: str, description: str, defect_type: DefectTypeModel
    ) -> ControlHandlerModel:
        """
        Create blank controller representation, should be
        populated with content separately (due to UI file upload util)
        """
        obj = ControlHandlerModel(
            title=title, description=description, defect_type=defect_type
        )
        self.session.add(obj)
        return obj

    def list_nested(self) -> List[ControlHandlerModel]:
        return (
            self.session.query(self._model)
            .options(joinedload(ControlHandlerModel.defect_type))
            .all()
        )

    def allow_delete_hook(self, id: int) -> bool:
        return not self.session.query(
            select(ControlTargetModel)
            .where(ControlTargetModel.control_handler_id == id)
            .exists()
        ).scalar()

    def post_delete_hook(self, obj: ControlHandlerModel):
        if obj.is_valid:
            obj.destroy_source()

    test_store_connection = _model.test_store_connection



import urllib3
from minio import Minio

from open_aoi.settings import MINIO_PORT, MINIO_ACCESS_KEY, MINIO_SECRET_KEY
from open_aoi.exceptions import ConnectivityError


class Mixin:
    id: int

    _client = Minio(
        f"127.0.0.1:{MINIO_PORT}",
        access_key=MINIO_ACCESS_KEY,
        secret_key=MINIO_SECRET_KEY,
        secure=False,
        http_client=urllib3.PoolManager(num_pools=10, timeout=10, retries=2),
    )

    @classmethod
    def test_store_connection(cls):
        try:
            client = cls._client
            client.bucket_exists("ignore")
        except Exception as e:
            raise ConnectivityError("Could not connect to source code storage.") from e



from open_aoi.models import ConnectedComponentModel, ControlZoneModel
from open_aoi.controllers import Controller


class ConnectedComponentController(Controller):
    _model = ConnectedComponentModel

    def create(
        self,
        stat_left: int,
        stat_top: int,
        stat_width: int,
        stat_height: int,
        control_zone: ControlZoneModel,
    ) -> ConnectedComponentModel:
        assert stat_left >= 0
        assert stat_top >= 0
        assert stat_width >= 0
        assert stat_height >= 0
        
        obj = ConnectedComponentModel(
            stat_left=stat_left,
            stat_top=stat_top,
            stat_width=stat_width,
            stat_height=stat_height,
            control_zone=control_zone,
        )
        self.session.add(obj)
        return obj

    def allow_delete_hook(self, id: int) -> bool:
        return True



from datetime import datetime
from typing import List, Optional

from sqlalchemy import select
from sqlalchemy.orm import joinedload

from open_aoi.models import (
    CameraModel,
    TemplateModel,
    AccessorModel,
    InspectionProfileModel,
    ControlZoneModel,
)
from open_aoi.controllers import Controller


class TemplateController(Controller):
    _model = TemplateModel

    def create(
        self, title: str, accessor: AccessorModel, image_blob: Optional[str] = None
    ) -> CameraModel:
        obj = TemplateModel(
            title=title,
            image_blob=image_blob,
            created_by=accessor,
            created_at=datetime.now(),
        )
        self.session.add(obj)
        return obj

    def allow_delete_hook(self, id: int) -> bool:
        any_inspection_profile = self.session.query(
            select(InspectionProfileModel)
            .where(InspectionProfileModel.template_id == id)
            .exists()
        ).scalar()
        any_control_zone = self.session.query(
            select(ControlZoneModel).where(ControlZoneModel.template_id == id).exists()
        ).scalar()
        return not (any_inspection_profile or any_control_zone)

    def list_nested(self) -> List[TemplateModel]:
        return (
            self.session.query(self._model)
            .options(joinedload(TemplateModel.control_zone_list))
            .options(joinedload(TemplateModel.inspection_profile_list))
            .all()
        )

    def post_delete_hook(self, obj: TemplateModel):
        if obj.is_valid:
            obj.destroy_image()

    test_store_connection = _model.test_store_connection



from datetime import datetime
from typing import Optional, List

from sqlalchemy import (
    String,
    ForeignKey,
    Numeric,
    DateTime,
    func,
    Boolean,
    Integer,
    MetaData,
)
from sqlalchemy import create_engine
from sqlalchemy.orm import Mapped, mapped_column, DeclarativeBase, relationship

from open_aoi.settings import MYSQL_DATABASE, MYSQL_PASSWORD, MYSQL_USER, MYSQL_PORT
from open_aoi.enums import RoleEnum, AccessorEnum
from open_aoi.mixins.authentication import AuthMixin, SessionCredentialsMixin
from open_aoi.mixins.image_source import (
    TemplateImageSourceMixin,
    InspectionImageSourceMixin,
)
from open_aoi.mixins.module_source import ModuleSourceMixin


CODE_LIMIT = 100
TITLE_LIMIT = 200
DESCRIPTION_LIMIT = 500


metadata_obj = MetaData()
engine = create_engine(
    f"mysql+pymysql://{MYSQL_USER}:{MYSQL_PASSWORD}@localhost:{MYSQL_PORT}/{MYSQL_DATABASE}"
)


class Base(DeclarativeBase):
    pass


class RoleModel(Base):
    """Define accessor rights"""

    __tablename__ = "Role"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(
        primary_key=True, nullable=False, autoincrement=True
    )

    allow_system_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_log_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_details_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_flow_control: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_accessor_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_device_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_profile_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_system_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_statistics_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )

    accessor_list: Mapped[List["AccessorModel"]] = relationship(back_populates="role")
    registry = RoleEnum


class AccessorModel(Base, AuthMixin, SessionCredentialsMixin):
    __tablename__ = "Accessor"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(
        primary_key=True, nullable=False, autoincrement=True
    )

    username: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)
    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)
    description: Mapped[str] = mapped_column(String(DESCRIPTION_LIMIT), nullable=False)

    role_id: Mapped[int] = mapped_column(ForeignKey("Role.id"), nullable=False)
    role: Mapped["RoleModel"] = relationship()

    hash: Mapped[str] = mapped_column(String(60), nullable=False)
    registry = AccessorEnum


class DefectTypeModel(Base):
    """Define system wide known error types"""

    __tablename__ = "DefectType"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)
    description: Mapped[str] = mapped_column(String(DESCRIPTION_LIMIT), nullable=False)

    control_handler_list: Mapped[List["ControlHandlerModel"]] = relationship(
        back_populates="defect_type"
    )


class ControlHandlerModel(Base, ModuleSourceMixin):
    """Database representation of control handler"""

    __tablename__ = "ControlHandler"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)
    description: Mapped[str] = mapped_column(String(DESCRIPTION_LIMIT), nullable=False)

    handler_blob: Mapped[str] = mapped_column(
        String(100), nullable=True
    )  # If null, should not be used!

    control_target_list: Mapped[List["ControlTargetModel"]] = relationship(
        back_populates="control_handler"
    )

    defect_type_id: Mapped[int] = mapped_column(
        ForeignKey("DefectType.id"), nullable=False
    )
    defect_type: Mapped["DefectTypeModel"] = relationship(
        back_populates="control_handler_list"
    )


class ControlTargetModel(Base):
    """
    Helper object to map control handler to the control zone (control handler is unique).
    Multiple control targets are allowed for single control zone.
    """

    __tablename__ = "ControlTarget"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    control_handler_id: Mapped[int] = mapped_column(
        ForeignKey("ControlHandler.id"), nullable=False
    )
    control_handler: Mapped["ControlHandlerModel"] = relationship(
        back_populates="control_target_list"
    )

    control_zone_id: Mapped[int] = mapped_column(
        ForeignKey("ControlZone.id"), nullable=False
    )
    control_zone: Mapped["ControlZoneModel"] = relationship(
        back_populates="control_target_list"
    )


class ConnectedComponentModel(Base):
    """
    Location description for control zone
    """

    __tablename__ = "ConnectedComponent"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    stat_left: Mapped[int] = mapped_column(Integer, nullable=False)
    stat_top: Mapped[int] = mapped_column(Integer, nullable=False)
    stat_width: Mapped[int] = mapped_column(Integer, nullable=False)
    stat_height: Mapped[int] = mapped_column(Integer, nullable=False)

    control_zone_id: Mapped[int] = mapped_column(
        ForeignKey("ControlZone.id"), nullable=False
    )
    control_zone: Mapped["ControlZoneModel"] = relationship(back_populates="cc")


class ControlZoneModel(Base):
    """
    Small zone on template where related control handler is applied in order to detect defect type
    """

    __tablename__ = "ControlZone"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)

    template_id: Mapped[int] = mapped_column(ForeignKey("Template.id"), nullable=False)
    template: Mapped["TemplateModel"] = relationship(back_populates="control_zone_list")

    cc: Mapped["ConnectedComponentModel"] = relationship(
        back_populates="control_zone", cascade="all, delete-orphan"
    )

    rotation: Mapped[float] = mapped_column(Numeric(precision=10, scale=2))

    control_target_list: Mapped[List["ControlTargetModel"]] = relationship(
        back_populates="control_zone",  # Cascade -prevent delete if any target use this control zone
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["AccessorModel"] = relationship()


class InspectionLogModel(Base):
    """
    Helper to map defect type that was found to control zone. Multiple control logs are allowed.
    """

    __tablename__ = "ControlLog"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    control_target_id: Mapped[int] = mapped_column(
        ForeignKey("ControlTarget.id"), nullable=False
    )
    control_target: Mapped["ControlTargetModel"] = relationship()
    passed: Mapped[bool] = mapped_column(Boolean(), default=False, nullable=False)
    log: Mapped[str] = mapped_column(String(200), nullable=True)
    inspection_id: Mapped[int] = mapped_column(
        ForeignKey("Inspection.id"), nullable=False
    )
    inspection: Mapped["InspectionModel"] = relationship(
        back_populates="inspection_log_list"
    )
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())


class InspectionModel(Base, InspectionImageSourceMixin):
    """
    Connect defect collection with template
    """

    __tablename__ = "Inspection"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    image_blob: Mapped[str] = mapped_column(String(100), nullable=False)

    inspection_profile_id: Mapped[int] = mapped_column(
        ForeignKey("InspectionProfile.id"), nullable=False
    )
    inspection_profile: Mapped["InspectionProfileModel"] = relationship(
        back_populates="inspection_list"
    )

    inspection_log_list: Mapped[List["InspectionLogModel"]] = relationship(
        back_populates="inspection"
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())

    @property
    def overall_passed(self):
        raise NotImplementedError()


class TemplateModel(Base, TemplateImageSourceMixin):
    """
    Main reference image. Aggregate control zones.
    """

    __tablename__ = "Template"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)

    image_blob: Mapped[str] = mapped_column(String(100), nullable=True)

    control_zone_list: Mapped[List["ControlZoneModel"]] = relationship(
        back_populates="template",  # Delete cascade - prevent if control zone use template
    )

    inspection_profile_list: Mapped["InspectionProfileModel"] = relationship(
        back_populates="template",  # Cascade should prevent delete operation if profiles are found
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["AccessorModel"] = relationship()


class CameraModel(Base):
    """
    Represent available cameras
    """

    __tablename__ = "Camera"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)
    description: Mapped[str] = mapped_column(String(DESCRIPTION_LIMIT), nullable=False)

    # TODO: insert validation ipv4
    ip_address: Mapped[str] = mapped_column(String(15), nullable=False)

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["AccessorModel"] = relationship()


class InspectionProfileModel(Base):
    """
    Concrete instance of desired test configuration
    """

    __tablename__ = "InspectionProfile"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    is_active: Mapped[bool] = mapped_column(Boolean, nullable=False, default=False)

    title: Mapped[str] = mapped_column(String(TITLE_LIMIT), nullable=False)
    description: Mapped[str] = mapped_column(String(DESCRIPTION_LIMIT), nullable=False)

    identification_code: Mapped[str] = mapped_column(String(CODE_LIMIT), nullable=False)

    camera_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Camera.id"), nullable=True
    )
    camera: Mapped[Optional["CameraModel"]] = relationship()

    template_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Template.id"), nullable=True
    )
    template: Mapped[Optional["TemplateModel"]] = relationship()

    inspection_list: Mapped[List["InspectionModel"]] = relationship(
        back_populates="inspection_profile"
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["AccessorModel"] = relationship()



from datetime import datetime
from sqlalchemy import select

from open_aoi.models import CameraModel, AccessorModel, InspectionProfileModel
from open_aoi.controllers import Controller


class CameraController(Controller):
    _model = CameraModel

    def create(
        self, title: str, description: str, ip_address: str, accessor: AccessorModel
    ) -> CameraModel:
        obj = CameraModel(
            title=title,
            description=description,
            ip_address=ip_address,
            created_by=accessor,
            created_at=datetime.now(),
        )
        self.session.add(obj)
        return obj

    def allow_delete_hook(self, id: int) -> bool:
        return not self.session.query(
            select(InspectionProfileModel)
            .where(InspectionProfileModel.camera_id == id)
            .exists()
        ).scalar()



"""Script create DB tables"""

from dotenv import load_dotenv

assert load_dotenv(".env")

from open_aoi.models import *

if __name__ == "__main__":
    metadata_obj.drop_all(engine)
    metadata_obj.create_all(engine)



from enum import Enum


class RoleEnum(Enum):
    """Supported role types (reflected in DB)"""

    OPERATOR = 1
    ADMINISTRATOR = 2


class AccessorEnum(Enum):
    """Supported accessors (reflected in DB)"""

    OPERATOR = 1
    ADMINISTRATOR = 2



from datetime import datetime
from typing import List, Optional

from sqlalchemy.orm import joinedload

from open_aoi.models import (
    InspectionProfileModel,
    InspectionModel,
)
from open_aoi.controllers import Controller


class InspectionController(Controller):
    _model = InspectionModel

    def create(
        self,
        inspection_profile: InspectionProfileModel,
        image_blob: Optional[str] = None,
    ) -> InspectionModel:
        obj = InspectionModel(
            inspection_profile=inspection_profile,
            image_blob=image_blob,
            created_at=datetime.now(),
        )
        self.session.add(obj)
        return obj

    def allow_delete_hook(self, id: int) -> bool:
        return False  # Prevent log delete

    def list_nested(self) -> List[InspectionModel]:
        return (
            self.session.query(self._model)
            .options(joinedload(InspectionModel.inspection_log_list))
            .options(joinedload(InspectionModel.inspection_profile))
            .all()
        )

    test_store_connection = _model.test_store_connection



from abc import ABC


class General(Exception, ABC):
    pass


class SecurityException(General, ABC):
    """Base exception for security issues"""


class AuthException(SecurityException):
    """Raised on auth attempt fail"""


class IntegrityError(General):
    """Raised if internal integrity was violated"""


class ConnectivityError(General):
    """Raised if connection to resource could not be established"""


class ROSServiceError(General):
    """Raised if ROS service fails"""



from sqlalchemy import select

from open_aoi.controllers import Controller
from open_aoi.models import (
    AccessorModel,
)


class AccessorController(Controller):
    _model = AccessorModel
    revoke_session_access = AccessorModel.revoke_session_access
    identify_session_accessor_id = AccessorModel.identify_session_access

    def retrieve_by_username(self, username: str) -> AccessorModel:
        q = select(self._model).where(self._model.username == username)
        res = self.session.scalars(q).one_or_none()
        return res

    def identify_session_accessor(self, storage: dict) -> AccessorModel:
        id = self._model.identify_session_access(storage)
        return self.retrieve(id)



import io
import logging
import numpy as np
from uuid import uuid4
from typing import Optional

from PIL import Image

from open_aoi.mixins import Mixin
from open_aoi.exceptions import IntegrityError

logger = logging.getLogger("mixin.image_source")


class ImageSourceMixin(Mixin):
    image: np.ndarray
    image_blob: Optional[str]

    _bucket_name: str

    @property
    def is_valid(self):
        return getattr(self, "image_blob", None) is not None

    def publish_image(self, im: Image):
        assert getattr(self, "image_blob", None) is None
        client = self._client

        if not client.bucket_exists(self._bucket_name):
            client.make_bucket(self._bucket_name)

        image_blob = str(uuid4())
        blob = io.BytesIO()

        im.save(blob, format="PNG")
        length = blob.tell()
        blob.seek(0)

        client.put_object(self._bucket_name, image_blob, blob, length)

        self.image_blob = image_blob

    def materialize_image(self) -> Image.Image:
        assert getattr(self, "image_blob") is not None
        client = self._client

        if not client.bucket_exists(self._bucket_name):
            raise IntegrityError("Image does not exist")

        obj = client.get_object(self._bucket_name, self.image_blob)
        im = Image.open(obj, formats=["PNG"])
        obj.close()

        return im

    def destroy_image(self):
        assert getattr(self, "image_blob") is not None
        client = self._client

        if not client.bucket_exists(self._bucket_name):
            raise IntegrityError("Image does not exist")

        client.remove_object(self._bucket_name, self.image_blob)

        self.image_blob = None


class InspectionImageSourceMixin(ImageSourceMixin):
    _bucket_name = "inspections"


class TemplateImageSourceMixin(ImageSourceMixin):
    _bucket_name = "templates"



from sqlalchemy import select

from open_aoi.controllers import Controller
from open_aoi.models import (
    ControlHandlerModel,
    ControlTargetModel,
    ControlZoneModel,
    InspectionLogModel,
)


class ControlTargetController(Controller):
    _model = ControlTargetModel

    def create(
        self,
        control_handler: ControlHandlerModel,
        control_zone: ControlZoneModel,
    ) -> ControlTargetModel:
        obj = ControlTargetModel(
            control_handler=control_handler, control_zone=control_zone
        )
        self.session.add(obj)
        return obj

    def allow_delete_hook(self, id: int) -> bool:
        return not self.session.query(
            select(InspectionLogModel)
            .where(InspectionLogModel.control_target_id == id)
            .exists()
        ).scalar()
